#include <referee.h>
#include <referee.moc>

#include <qdatetm.h>
#include <stdlib.h>
#include <qtimer.h>
#include <qcolor.h>
#include <qkeycode.h>

#if (KDE_VERSION_MAJOR >= 1 && KDE_VERSION_MINOR >= 1)
#include <kaccel.h>
#else
#include <kkeyconf.h>
#endif

#include "board.h"
#include "pacman.h"
#include "monster.h"
#include "fruit.h"
#include "painter.h"

Referee::Referee( QWidget *parent, const char *name, Bitfont *font, QString pixDir )
    : QWidget( parent, name )
{
    gameState.resize(11);
    gameTimer = 0;

    setFocusPolicy(QWidget::StrongFocus);

    initKeys();
    initTiming();
    initScoring();

    board = new Board(BoardWidth*BoardHeight);

    pix = new Painter(board, this, font, pixDir);
    setFixedSize(pix->levelPix().size());

    pacman = new Pacman(board);
    pacman->setMaxPixmaps(pix->maxPixmaps(PacmanPix));

    fruit = new Fruit(board);

    monsters = new QList<Monster>;
    monsters->setAutoDelete(TRUE);

    monsterRect = new QList<QRect>;
    monsterRect->setAutoDelete(TRUE);

    pacmanRect.setRect(0, 0, 0, 0);
    fruitRect.setRect(0, 0, 0, 0);

    QTime midnight( 0, 0, 0 );
    srand( midnight.secsTo(QTime::currentTime()) );

    lifes = 0;
    level = 0;
    points = 0;

    emit setLifes(lifes);
    emit setLevel(level);
    emit setPoints(points);

    intro();
}

void Referee::paintEvent( QPaintEvent *e)
{
    QRect rect = e->rect();

    if (!rect.isEmpty())
	bitBlt(this, rect.x(), rect.y(),
	       &pix->levelPix(), rect.x(), rect.y(), rect.width(), rect.height());

    if ((gameState.testBit(GameOver) || gameState.testBit(Demonstration)) &&
	rect.intersects(pix->rect(board->position(fruithome), i18n("GAME  OVER"))))
	pix->draw(board->position(fruithome), Widget, i18n("GAME  OVER"), RED);

    if (!gameState.testBit(Init)) {

	if (!gameState.testBit(Dying) &&
	    rect.intersects(pix->rect(fruit->position(), FruitPix)) && (fruit->pix() != -1))
	    if (fruit->pix() >= 0)
		if (fruit->state() != active)
		    switch (fruit->pix()) {
		        case 0  : pix->draw(fruit->position(), Widget, "#)", PINK);
			          break;
		        case 1  : pix->draw(fruit->position(), Widget, "%)", PINK);
			          break;
		        default : pix->draw(fruit->position(), Widget, "')", PINK);
			          break;
		    }
		else
		    pix->draw(fruit->position(), Widget, FruitPix, fruit->pix());

	for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	    if (m && m->state() == harmless && 
		rect.intersects(pix->rect(m->position(), MonsterPix)) &&
		!(m->position() == pacman->position() && gameState.testBit(Scoring))) {
	        if (m->body() != -1)
		    pix->draw(m->position(), Widget, MonsterPix, m->body());
		if (m->eyes() != -1)
		    pix->draw(m->position(), Widget, EyesPix, m->eyes());
	    }

	if (!gameState.testBit(Scoring) && !gameState.testBit(LevelDone) &&
	    rect.intersects(pix->rect(pacman->position(), PacmanPix)) && pacman->pix() != -1)
            pix->draw(pacman->position(), Widget, PacmanPix, pacman->pix());

	for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	    if (m && m->state() != harmless &&
		rect.intersects(pix->rect(m->position(), MonsterPix)) &&
		!(m->position() == pacman->position() && gameState.testBit(Scoring))) {
                if (m->body() != -1)
		    pix->draw(m->position(), Widget, MonsterPix, m->body());
		if (m->eyes() != -1)
		    pix->draw(m->position(), Widget, EyesPix, m->eyes());
	    }
    }

    if (gameState.testBit(Scoring) &&
	rect.intersects(pix->rect(pacman->position(), PacmanPix)))
	switch (monstersEaten) {
	    case 1 : pix->draw(pacman->position(), Widget, "$)", CYAN);
		     break;
	    case 2 : pix->draw(pacman->position(), Widget, "&)", CYAN);
		     break;
	    case 3 : pix->draw(pacman->position(), Widget, "()", CYAN);
		     break;
	    case 4 : pix->draw(pacman->position(), Widget, "*+", CYAN);
		     break;
	}

    if (gameState.testBit(Init) && gameState.testBit(Dying) &&
	timerCount < pix->maxPixmaps(DyingPix) &&
	rect.intersects(pix->rect(pacman->position(), PacmanPix)))
	pix->draw(pacman->position(), Widget, DyingPix, timerCount);

    if (gameState.testBit(LevelDone) &&
	rect.intersects(pix->rect(pacman->position(), PacmanPix)))
	pix->draw(pacman->position(), Widget, PacmanPix, 0);

    if (gameState.testBit(Player) &&
	rect.intersects(pix->rect(board->position(monsterhome, 0), i18n("PLAYER ONE"))))
	pix->draw(board->position(monsterhome, 0), Widget, i18n("PLAYER ONE"), CYAN);

    if (gameState.testBit(Ready) &&
	rect.intersects(pix->rect(board->position(fruithome), i18n("READY!"))))
	pix->draw(board->position(fruithome), Widget, i18n("READY!"), YELLOW);

    if (gameState.testBit(Paused) &&
	rect.intersects(pix->rect((BoardWidth*BoardHeight)/2-BoardWidth, i18n("PAUSED"))))
	pix->draw((BoardWidth*BoardHeight)/2-BoardWidth, Widget, i18n("PAUSED"), RED, BLACK);
}

void Referee::timerEvent( QTimerEvent * )
{
    timerCount++;

    QRect lastRect = pix->rect(pacman->position(), PacmanPix);
    int lastPix = pacman->pix();
    bool moved = FALSE;
    int eated = 0;

    if (moved = pacman->move()) {	// pacman really moved
	pacmanRect = pix->rect(pacman->position(), PacmanPix);
	if (lastPix == pacman->pix() &&
	    lastRect == pix->rect(pacman->position(), PacmanPix))
	    pacmanRect.setRect(0, 0, 0, 0);	// nothing to do, because the pixmap
	else					// and the position isn't changed.
	    pacmanRect = pacmanRect.unite(lastRect);
    } else
	pacmanRect.setRect(0, 0, 0, 0);

    int pos = pacman->position();

    if (moved && board->isMonster(pos) && !gameState.testBit(Dying)) {
	for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	    if (m  && m->position() == pos) {
		if (m->state() == harmless && !gameState.testBit(Dying)) {
		    m->setREM(remTicks[level]);
		    m->setDirection(X);	// prevent movement before eaten()
		    eated++;
		    if (gameState.testBit(Introducing))
		        m->setPosition(OUT);
		}
		if (m->state() == dangerous && !gameState.testBit(Dying))
		    killed();
	    }
    }

    if (moved && !gameState.testBit(Dying)) {
	if (board->isPoint(pos)) {
	    board->reset(pos, Point);
            score(pointScore);
	    pix->erase(pos, PointPix);
        }
        if (board->isSwitch(pos)) {
	    board->reset(pos, Switch);
            score(switchScore);
	    monstersEaten = 0;
	    pix->erase(pos, SwitchPix);
	    for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	        if (m && m->state() != rem) {
		    m->setHarmless(harmlessTicks[level], harmlessDurTicks[level],
 			           harmlessWarnTicks[level]);
		    if (gameState.testBit(Introducing))
			m->setDirection(board->turn(m->direction()));
		}
        }
	if (board->isFruit(pos) && fruit->state() == active) {
	    fruit->setEaten(fruitScoreDurTicks[level]);
	    score(fruitScore[fruit->pix()]);
	}
    }

    if (!gameState.testBit(Introducing)) {
	lastRect = pix->rect(fruit->position(), FruitPix);
	lastPix = fruit->pix();
	if (fruit->move()) {
	    fruitRect = pix->rect(fruit->position(), FruitPix);
	    if (lastPix == fruit->pix() &&
		lastRect == pix->rect(fruit->position(), FruitPix))
	        fruitRect.setRect(0, 0, 0, 0);
	    else
	        fruitRect = fruitRect.unite(lastRect);
	} else
	    fruitRect.setRect(0, 0, 0, 0);
    } else
	fruitRect.setRect(0, 0, 0, 0);

    int lastBodyPix;
    int lastEyesPix;

    for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	if (m) {
	    lastRect = pix->rect(m->position(), MonsterPix);
	    lastBodyPix = m->body();
	    lastEyesPix = m->eyes();
	    if (m->move()) {
		moved = TRUE;
		*monsterRect->at(m->id()) = pix->rect(m->position(), MonsterPix);
		if (lastBodyPix == m->body() && lastEyesPix == m->eyes() &&
	            lastRect == pix->rect(m->position(), MonsterPix))
		    monsterRect->at(m->id())->setRect(0, 0, 0, 0);
		else
		    *monsterRect->at(m->id()) = monsterRect->at(m->id())->unite(lastRect);
		if (m->position() == pos && !gameState.testBit(Dying)) {
		    if (m->state() == harmless && !gameState.testBit(Dying)) {
			m->setREM(remTicks[level]);
			eated++;
		        if (gameState.testBit(Introducing)) {
			    m->setPosition(OUT);
			    m->setDirection(X);
			}		
		    }
		    if (m->state() == dangerous && !gameState.testBit(Dying))
			killed();
		}
	    } else
		monsterRect->at(m->id())->setRect(0, 0, 0, 0);
	}

    for (int m = 0; m < board->monsters(); m++)
	if (pacmanRect.intersects(*monsterRect->at(m))) {
	    pacmanRect = pacmanRect.unite(*monsterRect->at(m));
	    monsterRect->at(m)->setRect(0, 0, 0, 0);
	} else
	    for (int im = m+1; im < board->monsters(); im++)
		if (monsterRect->at(m)->intersects(*monsterRect->at(im))) {
		    *monsterRect->at(m) = monsterRect->at(m)->unite(*monsterRect->at(im));
		    monsterRect->at(im)->setRect(0, 0, 0, 0);
		}
   
    if (!pacmanRect.isNull())
	repaint(pacmanRect, FALSE);

    if (!fruitRect.isNull())
	repaint(fruitRect, FALSE);

    for (int m = 0; m < board->monsters(); m++)
	if (!monsterRect->at(m)->isNull())
	    repaint(*monsterRect->at(m), FALSE);

    if (board->points() == 0 && !gameState.testBit(Dying))
	levelUp();

    if (eated > 0 && !gameState.testBit(Dying)) {
	timerCount = eated;
	eaten();
    }

    if (gameState.testBit(Introducing) && moved)
	introPlay();
}

void Referee::repaintFigures()
{
    pacmanRect = pix->rect(pacman->position(), PacmanPix);

    fruitRect = pix->rect(fruit->position(), FruitPix);

    for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	if (m)
	    *monsterRect->at(m->id()) = pix->rect(m->position(), MonsterPix);

    for (int m = 0; m < board->monsters(); m++)
	if (pacmanRect.intersects(*monsterRect->at(m))) {
	    pacmanRect = pacmanRect.unite(*monsterRect->at(m));
	    monsterRect->at(m)->setRect(0, 0, 0, 0);
	} else
	    for (int im = m+1; im < board->monsters(); im++)
		if (monsterRect->at(m)->intersects(*monsterRect->at(im))) {
		    *monsterRect->at(m) = monsterRect->at(m)->unite(*monsterRect->at(im));
		    monsterRect->at(im)->setRect(0, 0, 0, 0);
		}
    
    if (!pacmanRect.isNull())
	repaint(pacmanRect, FALSE);

    if (!fruitRect.isNull())
	repaint(fruitRect, FALSE);

    for (int m = 0; m < board->monsters(); m++)
	if (!monsterRect->at(m)->isNull())
	    repaint(*monsterRect->at(m), FALSE);

}

void Referee::setFont(Bitfont *font)
{
    pix->setFont(font);
}

void Referee::setPixmaps(QString pixDir)
{
    pix->setPixmaps(pixDir);
    pacman->setMaxPixmaps(pix->maxPixmaps(PacmanPix));

    for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	if (m)
	    m->setMaxPixmaps(pix->maxPixmaps(MonsterPix), pix->maxPixmaps(EyesPix));

    if (gameState.testBit(Introducing))
	for (int i = 0; i < (gameState.testBit(Init) ? timerCount : 15); i++)
	    introPaint(i);

    setFixedSize(pix->levelPix().size());
    repaint();
}

void Referee::initKeys()
{
    KConfig *conf = kapp->getConfig();

    QString up("Up");
    up = conf->readEntry("upKey", (const char*) up);
    UpKey    = stringToKey(up);

    QString down("Down");
    down = conf->readEntry("downKey", (const char*) down);
    DownKey  = stringToKey(down);

    QString left("Left");
    left = conf->readEntry("leftKey", (const char*) left);
    LeftKey  = stringToKey(left);

    QString right("Right");
    right = conf->readEntry("rightKey", (const char*) right);
    RightKey = stringToKey(right);

}

void fillArray(QArray<int> &array, QString values, int max)
{
    array.resize(max);
    int last = 0;
    bool ok;
    QString value;

    for (uint i = 0; i < array.size(); i++) {
	if (values.find(',') < 0 &&  values.length() > 0) {
	    value = values;
	    values = "";
        }
	if (values.find(',') >= 0) {
	    value = values.left(values.find(','));
	    values.remove(0,values.find(',')+1);
	}
	array[i] = value.toInt(&ok);
	if (ok)
	    last = array[i];
	else
            array[i] = last;
    }
}

void Referee::initTiming()
{
    KConfig *conf = kapp->getConfig();
    
    maxLevel = conf->readNumEntry("Levels", 7);

    fillArray(speed, conf->readEntry("SpeedMS", "20"), maxLevel+1);
    fillArray(monsterIQ, conf->readEntry("MonsterIQ", "0,170,180,170,180,170,180"), maxLevel+1);
    fillArray(pacmanTicks,conf->readEntry("PacmanTicks", "3"), maxLevel+1);
    fillArray(remTicks, conf->readEntry("RemTicks", "1"), maxLevel+1);
    fillArray(dangerousTicks, conf->readEntry("DangerousTicks", "3,,,,,,2,3,2"), maxLevel+1);
    fillArray(harmlessTicks, conf->readEntry("HarmlessTicks", "7,6,,5,,4"), maxLevel+1);
    fillArray(harmlessDurTicks, conf->readEntry("HarmlessDurationTicks", "375,,,300,,250,200,150"), maxLevel+1);
    fillArray(harmlessWarnTicks, conf->readEntry("HarmlessWarningTicks", "135"), maxLevel+1);
    fillArray(arrestTicks, conf->readEntry("ArrestTicks", "6"), maxLevel+1);
    fillArray(arrestDurTicks, conf->readEntry("ArrestDurationTicks", "200,,,150"), maxLevel+1);
    fillArray(fruitAppearsTicks, conf->readEntry("FruitAppearsTicks", "1000,,1500,2000,2500,3000,3500,4000"), maxLevel+1);
    fillArray(fruitDurTicks, conf->readEntry("FruitDurationTicks", "500,,,400,350,300,,250,200,150"), maxLevel+1);
    fillArray(fruitScoreDurTicks, conf->readEntry("FruitScoreDurationTicks", "150"), maxLevel+1);

    monsterScoreDurMS = conf->readNumEntry("MonsterScoreDurationMS", 1000);
    playerDurMS = conf->readNumEntry("PlayerDurationMS", 3000);
    readyDurMS = conf->readNumEntry("ReadyDurationMS", 2000);
    gameOverDurMS = conf->readNumEntry("GameOverDurationMS", 3000);
    afterPauseMS = conf->readNumEntry("AfterPauseMS", 1000);
    dyingPreAnimationMS = conf->readNumEntry("DyingPreAnimationMS", 1000); 
    dyingAnimationMS = conf->readNumEntry("DyingAnimationMS", 100);
    dyingPostAnimationMS = conf->readNumEntry("DyingPostAnimationMS", 500);
    introAnimationMS = conf->readNumEntry("IntroAnimationMS", 800);
    introPostAnimationMS = conf->readNumEntry("IntroPostAnimationMS", 1000);
    levelUpPreAnimationMS = conf->readNumEntry("LevelUpPreAnimationMS", 2000); 
    levelUpAnimationMS = conf->readNumEntry("LevelUpAnimationMS", 2000);
}

void Referee::initScoring()
{
    KConfig *conf = kapp->getConfig();

    pointScore = conf->readNumEntry("PointScore", 10);
    switchScore = conf->readNumEntry("SwitchScore", 50);
    fillArray(fruitScore, conf->readEntry("FruitScore", "100,300,500"), maxLevel);
    fillArray(monsterScore, conf->readEntry("MonsterScore", "200,400,800,1600"), 4);
    extraLifeScore = conf->readNumEntry("ExtraLifeScore", 10000);
}

void Referee::keyPressEvent( QKeyEvent *k )
{
    if (gameState.testBit(Paused) || gameState.testBit(Demonstration) ||
	gameState.testBit(Dying))
	return;

    uint key = k->key();
    if (key == UpKey)
	pacman->setDirection(N);
    else if (key == DownKey)
	pacman->setDirection(S);
    else if (key == RightKey)
	pacman->setDirection(E);
    else if (key == LeftKey)
	pacman->setDirection(W);
    else {
	k->ignore();
	return;
    }
    k->accept();
}

void Referee::score(int p)
{
    if (!gameState.testBit(Playing))
	return;

    int lastPoints = points;

    points += p;
    points = (points < 0 ? 0 : points);
    emit setPoints(points);

    if ((int) lastPoints/extraLifeScore != (int) points/extraLifeScore)
	emit setLifes(++lifes);
}

void Referee::eaten()
{
    if (gameState.testBit(Ready))
	return;

    stop();
    
    if (monstersEaten < 4)
	monstersEaten++;

    gameState.setBit(Scoring);
    score(monsterScore[monstersEaten-1]);

    repaint(pix->rect(pacman->position(), PacmanPix));

    if (--timerCount > 0)
        QTimer::singleShot( monsterScoreDurMS, this, SLOT(eaten()));
    else {
	for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	    if (m && m->direction() == X && !gameState.testBit(Introducing))
	        m->setDirection(N);
	if (monstersEaten != 4 || !gameState.testBit(Introducing))
            QTimer::singleShot( monsterScoreDurMS, this, SLOT(start()));
    }
}

void Referee::pause()
{
    static int pausedTimer = 0;

    if (!gameState.testBit(Paused)) {
	pausedTimer = gameTimer;
        stop();
	gameState.setBit(Paused);
	repaint(pix->rect((BoardWidth*BoardHeight)/2-BoardWidth, i18n("PAUSED")), FALSE);
    } else {
	gameState.clearBit(Paused);
	repaint(pix->rect((BoardWidth*BoardHeight)/2-BoardWidth, i18n("PAUSED")), FALSE);
	if (pausedTimer) {
	    pausedTimer = 0;
	    start();
	}
    }
    emit togglePaused();
}

void Referee::intro()
{
    if (gameState.testBit(Ready))
	return;

    stop();
    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(intro()));
	return;
    }

    gameState.fill(FALSE);
    gameState.setBit(Introducing);
    gameState.setBit(Init);

    level = 0;
    emit setLevel(level);

    board->init(Intro);
    pix->initRoomPixmap();

    initPacman();
    initFruit();
    initMonsters();

    repaint();

    monstersEaten = 0;
    timerCount = 0;
    introPlay();
}

void Referee::introMonster(int id)
{
    Monster *m = new Monster(board, id);

    m->setPosition((10+id*6)*BoardWidth+10);
    m->setDirection(E);
    m->setDangerous(dangerousTicks[level], monsterIQ[level]);
    m->setMaxPixmaps(pix->maxPixmaps(MonsterPix), pix->maxPixmaps(EyesPix));

    if (m->body() != -1)
	pix->draw(m->position(), RoomPix, MonsterPix, m->body());
    if (m->eyes() != -1)
	pix->draw(m->position(), RoomPix, EyesPix, m->eyes());

    repaint(pix->rect(m->position(), MonsterPix), FALSE);
    m->setPosition(OUT);
}

void Referee::introPaint(int t)
{
    switch (t) {
	case  0 : repaint(pix->draw(35, 6, RoomPix, "CHARACTER / NICKNAME", WHITE), FALSE);
	 	  break;
	case  1 : introMonster(0);
		  break;
	case  2 : repaint(pix->draw(15+7, 10, RoomPix, "-SHADOW", RED), FALSE);
		  break;
	case  3 : repaint(pix->draw(37+8, 10, RoomPix, "\"BLINKY\"", RED), FALSE);
		  break;
	case  4 : introMonster(1);
		  break;
	case  5 : repaint(pix->draw(15+7, 16, RoomPix, "-SPEEDY", PINK), FALSE);
		  break;
	case  6 : repaint(pix->draw(37+7, 16, RoomPix, "\"PINKY\"", PINK), FALSE);
		  break;
	case  7 : introMonster(2);
		  break;
	case  8 : repaint(pix->draw(15+8, 22, RoomPix, "-BASHFUL", CYAN), FALSE);
		  break;
	case  9 : repaint(pix->draw(37+6, 22, RoomPix, "\"INKY\"", CYAN), FALSE);
		  break;
	case 10 : introMonster(3);
		  break;
	case 11 : repaint(pix->draw(15+6, 28, RoomPix, "-POKEY", ORANGE), FALSE);
		  break;
	case 12 : repaint(pix->draw(37+7, 28, RoomPix, "\"CLYDE\"", ORANGE), FALSE);
		  break;
	case 13 : repaint(pix->draw(25+2, 44, RoomPix, "10", WHITE), FALSE);
	          repaint(pix->draw(30+3, 44, RoomPix, "<=>", WHITE), FALSE);
	          repaint(pix->draw(25+2, 48, RoomPix, "50", WHITE), FALSE);
	          repaint(pix->draw(30+3, 48, RoomPix, "<=>", WHITE), FALSE);
		  break;
	case 14 : // "@ 1980 MIDWAY MFG.CO."
                  repaint(pix->draw(10+20, 58, RoomPix, "@ 1998 JOE THOENNISSEN", PINK), FALSE);
		  break;
    }
}

void Referee::introPlay()
{
    if (!gameState.testBit(Introducing) || gameState.testBit(Ready))
	return;
    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(introPlay()));
	return;
    }

    if (!gameState.testBit(Init)) {
	if (monstersEaten == 4) {
	    stop();
	    QTimer::singleShot(introPostAnimationMS, this, SLOT(demo()));
	}
	if (pacman->direction() == W) {
	    int id = -1;
            if (pacman->position() == 37*BoardWidth-6)
	        id = 0;
	    else
	        if (board->isMonster(37*BoardWidth-6))
		    for (Monster *m = monsters->first(); m != 0; m = monsters->next())
		        if (m && m->position() == 37*BoardWidth-6) {
			    id = m->id();
			    id++;
			    break;	
		        }
		
	    if (id >= 0 && id <= 4)
		for (Monster *m = monsters->first(); m != 0; m = monsters->next())
		    if (m && m->id() == id && m->position() == OUT) {
		        m->setPosition(37*BoardWidth-1);
		        m->setDirection(W);
		        m->setDangerous(dangerousTicks[level], monsterIQ[level]);
		        board->set(37*BoardWidth-1, monsterhome, id);
		        repaint(pix->rect(m->position(), MonsterPix));
		        break;
		    }
	}
	return;
    }

    if (timerCount < 15)
	introPaint(timerCount);

    switch (timerCount) {
	case 13 : board->set(44*BoardWidth+22, Point);
		  pix->drawPoint(44*BoardWidth+22);
		  repaint(pix->rect(44*BoardWidth+22, PointPix), FALSE);
		  board->set(48*BoardWidth+22, Switch);
		  pix->drawSwitch(48*BoardWidth+22);
		  repaint(pix->rect(48*BoardWidth+22, SwitchPix), FALSE);
		  break;
	case 14 : board->set(36*BoardWidth+10, Switch);
		  pix->drawSwitch(36*BoardWidth+10);
		  repaint(pix->rect(36*BoardWidth+10, SwitchPix), FALSE);
		  for (int pos = 8; pos < BoardWidth; pos++) {
		      board->set(34*BoardWidth+pos, out);
		      board->set(38*BoardWidth+pos, out);
		  }
		  board->set(36*BoardWidth+8, out);
		  break;
	case 15 : gameState.clearBit(Init);
		  initPacman();
    		  pacman->setDemo(TRUE);
		  pacman->setPosition(37*BoardWidth-1);
		  repaintFigures();
		  start();
		  return;		  
    }

    if (timerCount++ < 15)
	QTimer::singleShot(introAnimationMS, this, SLOT(introPlay()));
}

void Referee::demo()
{
    if (gameState.testBit(Ready))
	return;

    stop();

    gameState.fill(FALSE);
    gameState.setBit(Init);
    gameState.setBit(Demonstration);

    board->init(Demo);
    pix->initRoomPixmap();

    level = 0;
    emit setLevel(level);

    initPacman();
    initFruit();
    initMonsters();

    gameState.clearBit(Init);

    repaint();

    timerCount = 0;
    QTimer::singleShot(playerDurMS, this, SLOT(start()));
}

void Referee::play()
{
    if (gameState.testBit(Playing))
	return;

    stop();

    if (gameState.testBit(Paused)) {
	pause();
	stop();
    }

    gameState.fill(FALSE);
    gameState.setBit(Init);
    gameState.setBit(Playing);
    gameState.setBit(Player);
    gameState.setBit(Ready);

    lifes = 3;
    level = 1;
    points = 0;

    board->init(Level);
    pix->initRoomPixmap();

    initPacman();
    initFruit();
    initMonsters();

    repaint();
    emit setLifes(lifes);
    emit setLevel(level);
    emit setPoints(points);

    repaint(pix->rect(board->position(monsterhome, 0), i18n("PLAYER ONE")), FALSE);
    repaint(pix->rect(board->position(fruithome), i18n("READY!")), FALSE);

    timerCount = 0;
    QTimer::singleShot(playerDurMS, this, SLOT(ready()));
}

void Referee::ready()
{
    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(ready()));
	return;
    }

    if (gameState.testBit(Player)) {
	emit setLifes(--lifes);
	gameState.clearBit(Player);
	gameState.clearBit(Init);
	repaint(pix->rect(board->position(monsterhome, 0), i18n("PLAYER ONE")), FALSE);
	repaintFigures();
	QTimer::singleShot(playerDurMS, this, SLOT(ready()));
	return;
    }

    if (gameState.testBit(Ready)) {
	gameState.clearBit(Ready);
	repaint(pix->rect(board->position(fruithome), i18n("READY!")), FALSE);
	start();
    } else {
	gameState.setBit(Ready);
	gameState.clearBit(Init);
	repaint(pix->rect(board->position(fruithome), i18n("READY!")), FALSE);
	QTimer::singleShot(readyDurMS, this, SLOT(ready()));
    }
}


void Referee::levelUp()
{
    stop();

    gameState.setBit(LevelDone);
    repaint(pix->rect(pacman->position(), PacmanPix));
    
    timerCount = 0;
    QTimer::singleShot(levelUpPreAnimationMS, this, SLOT(levelUpPlay()));
}

void Referee::levelUpPlay()
{
    if (gameState.testBit(Ready))
	return;

    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(levelUpPlay()));
	return;
    }

    switch (timerCount) {
	case 0 : gameState.setBit(Init);
	         repaintFigures();
		 break;
	case 1 : gameState.clearBit(LevelDone);
		 repaint(pix->rect(pacman->position(), PacmanPix));
		 break;
    }
   
    if (timerCount++ < 2) {
	QTimer::singleShot(levelUpAnimationMS, this, SLOT(levelUpPlay()));
	return;
    }

    gameState.clearBit(Init);

    if (level < maxLevel)
	level++;

    board->init(Level);
    pix->initRoomPixmap();

    initPacman();
    initFruit();
    initMonsters();

    repaint();
    emit setLevel(level);

    ready();
}

void Referee::start()
{
    if (gameState.testBit(Ready))
	return;

    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(start()));
	return;
    }
    gameState.clearBit(Scoring);
	
    if (!gameTimer)
	gameTimer = startTimer( speed [level] );
}

void Referee::start(int t)
{
    gameTimer = startTimer(t);
}

void Referee::stop()
{
    if (gameTimer) {
    	killTimer (gameTimer);
	gameTimer = 0;
    }
}

void Referee::killed()
{
    if (gameState.testBit(Ready))
	return;

    if (!gameState.testBit(Dying)) {
	gameState.setBit(Dying);

	pacman->setDirection(X, TRUE);
	for (Monster *m = monsters->first(); m != 0; m = monsters->next())
	    if (m)
		m->setDirection(X);
	QTimer::singleShot(dyingPreAnimationMS, this, SLOT(killed()));
    } else {
	stop();
	if (gameState.testBit(Paused)) {
	    QTimer::singleShot(afterPauseMS, this, SLOT(killed()));
	    return;
	}

	gameState.setBit(Init);

	repaintFigures();

	timerCount = 0;
	killedPlay();
    }
}

void Referee::killedPlay()
{
    if (!gameState.testBit(Dying) || gameState.testBit(Ready))
	return;
    if (gameState.testBit(Paused)) {
	QTimer::singleShot(afterPauseMS, this, SLOT(killedPlay()));
	return;
    }

    if (timerCount <= pix->maxPixmaps(DyingPix)) {
	repaint(pix->rect(pacman->position(), PacmanPix), FALSE);
	if (timerCount >= pix->maxPixmaps(DyingPix)-1 || timerCount == 0)
	    QTimer::singleShot(dyingPostAnimationMS, this, SLOT(killedPlay()));
	else
	    QTimer::singleShot(dyingAnimationMS, this, SLOT(killedPlay()));
	timerCount++;
    } else {
	gameState.clearBit(Dying);
	if (lifes == 0) {
	    gameState.setBit(GameOver);
	    repaint(pix->rect(board->position(fruithome), i18n("GAME  OVER")), FALSE);
	    QTimer::singleShot(gameOverDurMS, this, SLOT(intro()));
	} else {
	    gameState.clearBit(Init);
	    initPacman();
	    initFruit();
	    initMonsters();
	    emit setLifes(--lifes);
	    repaintFigures(); 
            ready();
	}
    }
}

void Referee::initPacman()
{
    pacman->setDemo(gameState.testBit(Demonstration));
    pacman->setPosition(board->position(pacmanhome));
    pacman->setDirection(W, TRUE);
    pacman->setAlive(pacmanTicks[level]);
}

void Referee::initFruit()
{
    fruit->setLevel(level, fruitAppearsTicks[level], fruitDurTicks[level] );
}

void Referee::initMonsters()
{
    if( !monsters->isEmpty())
	monsters->clear();
    if( !monsterRect->isEmpty())
	monsterRect->clear();

    for (int id = 0; id < (gameState.testBit(Introducing) ? 4 : board->monsters()); id++) {
	Monster *m = new Monster(board, id);
	monsters->append(m);
	QRect *r = new QRect();
	monsterRect->append(r);
	if (!gameState.testBit(Introducing)) {
            m->setFreedom(board->position(prisonexit));
	    m->setDangerous(dangerousTicks[level], monsterIQ[level]);
	    if (id == 0)
                m->setPrison(board->position(prisonentry));
	    else {
                m->setPrison(board->position(monsterhome, id));
	        m->setArrested(arrestTicks[level], arrestDurTicks[level]*id);
	    }
	    m->setPosition(board->position(monsterhome, id));
	    switch(id) {
	        case 0  : m->setDirection(W); break;
	        case 1  : m->setDirection(N); break;
	        default : m->setDirection(S);
	    }
	}
	m->setMaxPixmaps(pix->maxPixmaps(MonsterPix), pix->maxPixmaps(EyesPix));
    }
}

void Referee::reloadRoomPixmap()
{
    if (gameState != Introducing) {
	pix->initbackPixmaps();
	pix->initRoomPixmap();
	repaint();
    }
}

